//////////////////////////////////////////////////////////////////////////////////
// Author: lsilvest
// 
// Create Date:    02/02/2008 
//
// Module Name:    main_00
// 
// Target Devices: Spartan 3E starter kit Rev D
//
// Tool versions:  Tested with ISE WebPack 9.2
//
// Description: This module does a read or a write to DDR by generating the
//              correct signals for the MIG DDR module.
//
////////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2008 Authors
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
/////////////////////////////////////////////////////////////////////////////////

module ddr_read_write
  (input         clk0,
   input         clk90,
   input         sys_rst,
   input         sys_rst90,
   input         sys_rst180,

   input         user_init_done,
   input  [31:0] user_output_data,
   output [31:0] user_input_data,
   output [25:0] user_input_address,
   output  [2:0] user_command,
   input         user_cmd_ack,
   input         user_data_valid,
   output        user_burst_done,

   input  [31:0] wr_input_data,
   input  [25:0] wr_address,
   output [31:0] wr_output_data,
   input wr_read,
   input wr_doit,
   output wr_done,
   output [7:0] leds
   );

  // user commands:                      
  parameter NOP_CMD        = 3'b000,
              PRECHARGE_CMD  = 3'b001,
              INIT_CMD       = 3'b010,
              WRITE_CMD      = 3'b100,
              READ_CMD       = 3'b110;

  parameter INIT                  = 4'b0000,
              WAIT_INIT           = 4'b0001,
              READY               = 4'b0010,
              WAIT_3CLK           = 4'b0011,
              BURST_DONE          = 4'b0100,
              READ_DATA           = 4'b0101,
              WAIT_USER_CMD_ACK_N = 4'b0110;
  
  
  reg [3:0] state;
  reg [1:0] wait_cntr;
  
  
  reg [31:0] user_input_data_r;
  reg [25:0] user_input_address_r;
  reg [2:0]  user_command_r;
  reg        user_burst_done_r; 
  reg        wr_done_r;
  reg [31:0] wr_output_data_r;
  reg        led_r;
  

  assign user_input_data = user_input_data_r;
  assign user_input_address = user_input_address_r;
  assign user_command = user_command_r; 
  assign user_burst_done = user_burst_done_r;   
  assign wr_done = wr_done_r;
  assign wr_output_data = wr_output_data_r;
  assign leds[3:0] = state;
  assign leds[7] = user_init_done;
  
  initial begin
    user_command_r <= NOP_CMD;
    user_burst_done_r <= 1'b0;
    state <= INIT;
    wr_done_r <= 1'b0;
    wait_cntr <= 1'b0;
  end

  
  // trivial processing of the various MIG 2.0 DDR SDRAM interface
  // signals: we take advantage of BL==2, so there is only
  // ever one 32-bit value to write and read.
  always@ (negedge clk0) begin
    case (state)

      INIT:
        if (!sys_rst && !sys_rst90 && !sys_rst180) begin
          user_command_r <= INIT_CMD;
          state <= WAIT_INIT;
        end

      WAIT_INIT:
        begin
          user_command_r <= NOP_CMD;
          if (user_init_done) begin
            state <= READY;
          end
        end

      READY:
        if (wr_doit) begin
          wr_done_r <= 1'b0;
          if (!wr_read) begin
            user_command_r <= WRITE_CMD;
            // normally the user input data is put on posedge clk90, but it doesn't
            // matter if it's earlier:
            user_input_data_r <= wr_input_data;
          end else begin
            user_command_r <= READ_CMD;
          end
          user_input_address_r <= wr_address;
          wait_cntr <= 1'b0;
          state <= WAIT_3CLK;
        end

      WAIT_3CLK:
        if (user_cmd_ack) begin
          if (wait_cntr < 2'b11) begin
            wait_cntr <= wait_cntr + 2'b1;
          end else begin
            user_burst_done_r <= 1'b1;
            user_command_r <= NOP_CMD;
            state <= WAIT_USER_CMD_ACK_N;
          end
        end

      WAIT_USER_CMD_ACK_N:
        begin
          user_burst_done_r <= 1'b0;
          if (!user_cmd_ack) begin
            wr_done_r <= 1'b1;
            state <= READY;
          end
        end
    endcase // case (state)
    //   
  end // always@ (negedge clk0)

  
  always@ (posedge clk90) begin
    if (user_data_valid) begin
      wr_output_data_r <= user_output_data;
    end
  end
  
endmodule // ddr_read_write
